# Generated by Django 1.9.1 on 2016-10-11 18:11
from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="File",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(db_index=True, max_length=255, unique=True)),
                ("size", models.PositiveIntegerField(db_index=True)),
                ("content", models.BinaryField()),
                (
                    "created_datetime",
                    models.DateTimeField(
                        db_index=True,
                        default=django.utils.timezone.now,
                        verbose_name="Created datetime",
                    ),
                ),
                (
                    "_content_hash",
                    models.CharField(
                        blank=True,
                        db_column="content_hash",
                        db_index=True,
                        max_length=128,
                        null=True,
                    ),
                ),
            ],
            options={
                "db_table": "binary_database_files_file",
            },
        ),
    ]



from django.conf import settings
from django.db import models
from django.utils import timezone

from django.db.models import BinaryField

from binary_database_files import utils
from binary_database_files.utils import write_file, is_fresh
from binary_database_files.manager import FileManager


class File(models.Model):

    objects = FileManager()

    name = models.CharField(
        max_length=255, unique=True, blank=False, null=False, db_index=True
    )

    size = models.PositiveIntegerField(db_index=True, blank=False, null=False)

    content = BinaryField(blank=False, null=False)

    created_datetime = models.DateTimeField(
        db_index=True, default=timezone.now, verbose_name="Created datetime"
    )

    _content_hash = models.CharField(
        db_column="content_hash", db_index=True, max_length=128, blank=True, null=True
    )

    class Meta:
        db_table = "binary_database_files_file"

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):

        # Check for and clear old content hash.
        if self.id:
            old = File.objects.get(id=self.id)
            if old.content != self.content:
                self._content_hash = None

        # Recalculate new content hash.
        self.content_hash

        return super(File, self).save(*args, **kwargs)

    @property
    def content_hash(self):
        if not self._content_hash and self.content:
            self._content_hash = utils.get_text_hash(self.content)
        return self._content_hash

    def dump(self, check_hash=False):
        """
        Writes the file content to the filesystem.

        If check_hash is true, clears the stored file hash and recalculates.
        """
        if is_fresh(self.name, self._content_hash):
            return
        write_file(self.name, self.content, overwrite=True)
        if check_hash:
            self._content_hash = None
        self.save()

    @classmethod
    def dump_files(cls, debug=True, verbose=False):
        """
        Writes all files to the filesystem.
        """
        if debug:
            tmp_debug = settings.DEBUG
            settings.DEBUG = False
        try:
            q = cls.objects.only("id", "name", "_content_hash").values_list(
                "id", "name", "_content_hash"
            )
            total = q.count()
            if verbose:
                print("Checking %i total files..." % (total,))
            i = 0
            for (file_id, name, content_hash) in q.iterator():
                i += 1
                if verbose and not i % 100:
                    print("%i of %i" % (i, total))
                if not is_fresh(name=name, content_hash=content_hash):
                    if verbose:
                        print(
                            ("File %i-%s is stale. Writing to local file " "system...")
                            % (file_id, name)
                        )
                    f = File.objects.get(id=file_id)
                    write_file(f.name, f.content, overwrite=True)
                    f._content_hash = None
                    f.save()
        finally:
            if debug:
                settings.DEBUG = tmp_debug



from django.core.management.base import BaseCommand

from binary_database_files.models import File


class Command(BaseCommand):
    help = (
        "Dumps all files in the database referenced by FileFields "
        "or ImageFields onto the filesystem in the directory specified by "
        "MEDIA_ROOT."
    )

    def handle(self, *args, **options):
        File.dump_files(verbose=True)



from django.conf import settings
from django.core.management.base import BaseCommand

from binary_database_files.models import File


class Command(BaseCommand):
    args = "<filename 1> <filename 2> ... <filename N>"
    help = (
        "Regenerates hashes for files. If no filenames given, " + "rehashes everything."
    )

    def handle(self, *args, **options):
        tmp_debug = settings.DEBUG
        settings.DEBUG = False
        try:
            q = File.objects.all()
            if args:
                q = q.filter(name__in=args)
            total = q.count()
            i = 1
            for f in q.iterator():
                print("%i of %i: %s" % (i, total, f.name))
                f._content_hash = None
                f.save()
                i += 1
        finally:
            settings.DEBUG = tmp_debug



"""Custom storage backend that stores files in the database to facilitate scaling."""
import os
from io import UnsupportedOperation, BytesIO

from django.conf import settings
from django.core import files
from django.core.files.storage import FileSystemStorage
from django.utils._os import safe_join

from binary_database_files import models
from binary_database_files import utils
from binary_database_files import settings as _settings


class DatabaseFile(files.File):
    def close(self):
        pass


class DatabaseStorage(FileSystemStorage):
    """Subclass of FileSystemStorage that implements the necessary methods to use the database for files."""

    def __init__(self, *args, **kwargs):
        """Override __init__ to allow passing of base_url parameter to be able to resolve to absolute URL."""
        super(DatabaseStorage, self).__init__(*args, **kwargs)
        # we check first if base_url has been passed to __init__, if not we try settings and finally fall back to ''
        self._base_url = (
            kwargs.get("base_url") or settings.DATABASE_FILES_BASE_URL
            if hasattr(settings, "DATABASE_FILES_BASE_URL")
            else ""
        )

    def _open(self, name, mode="rb"):
        """Open file with filename `name` from the database."""
        name = self.get_instance_name(name)
        try:
            # Load file from database.
            f = models.File.objects.get_from_name(name)
            content = f.content
            size = f.size
            if _settings.DB_FILES_AUTO_EXPORT_DB_TO_FS and not utils.is_fresh(
                f.name, f.content_hash
            ):
                # Automatically write the file to the filesystem
                # if it's missing and exists in the database.
                # This happens if we're using multiple web servers connected
                # to a common database behind a load balancer.
                # One user might upload a file from one web server, and then
                # another might access if from another server.
                utils.write_file(name, f.content)
        except models.File.DoesNotExist:
            # If not yet in the database, check the local file system
            # and load it into the database if present.
            fqfn = self.path(name)
            if os.path.isfile(fqfn):
                # print('Loading file into database.')
                self._save(name, open(fqfn, mode))
                fh = super(DatabaseStorage, self)._open(name, mode)
                content = fh.read()
                size = fh.size
            else:
                # Otherwise we don't know where the file is so we return an
                # empty file
                size = 0
                content = b""
        # Normalize the content to a new file object.
        fh = BytesIO(content)
        fh.name = name
        fh.mode = mode
        fh.size = size
        o = DatabaseFile(fh)
        return o

    def _save(self, name, content):
        """Save file with filename `name` and given content to the database."""
        name = self.get_instance_name(name)
        # ZipExtFile advertises seek() but can raise UnsupportedOperation
        try:
            content.seek(0)
        except UnsupportedOperation:
            pass
        content = content.read()
        if isinstance(content, str):
            content = content.encode("utf-8")
        size = len(content)
        models.File.objects.create(
            content=content,
            size=size,
            name=name,
        )
        # Automatically write the change to the local file system.
        if _settings.DB_FILES_AUTO_EXPORT_DB_TO_FS:
            utils.write_file(name, content, overwrite=True)
        return name

    def get_instance_name(self, name):
        """
        Normalize the name relative to the MEDIA_ROOT so that we extract files to the correct
        location under MEDIA_ROOT based on the File model instance, without needing to
        have access to the Storage instance, e.g. so that File.dump_files() can work.
        """
        root_path = os.path.abspath(settings.MEDIA_ROOT)
        assert self.location.startswith(root_path)  # noqa: S101
        relative_location = self.location[len(root_path) + 1 :]
        if (
            relative_location
            and name[: len(relative_location) + 1] == relative_location + os.path.sep
        ):
            # Name already normalized to the media root
            return name
        full_path = safe_join(self.location, name)
        return full_path[len(root_path) + 1 :]

    def _path(self, instance_name):
        return safe_join(settings.MEDIA_ROOT, instance_name)

    def path(self, name):
        """
        Return a local filesystem path where the file can be retrieved using
        Python's built-in open() function.

        File names are normalized to the MEDIA_ROOT.

        If the file has not been saved to disk, a NotImplementedError will
        be raised.
        """
        localpath = self._path(self.get_instance_name(name))
        if not os.path.exists(localpath):
            raise NotImplementedError
        return localpath

    def exists(self, name):
        """Return True if a file with the given filename exists in the database. Return False otherwise."""
        name = self.get_instance_name(name)
        if models.File.objects.filter(name=name).exists():
            return True
        localpath = self._path(name)
        return os.path.exists(localpath)

    def delete(self, name):
        """Delete the file with filename `name` from the database and filesystem."""
        name = self.get_instance_name(name)
        try:
            models.File.objects.get_from_name(name).delete()
            hash_fn = utils.get_hash_fn(name)
            if os.path.isfile(hash_fn):
                os.remove(hash_fn)
        except models.File.DoesNotExist:
            pass
        localpath = self._path(name)
        if os.path.exists(localpath):
            return super(DatabaseStorage, self).delete(name)

    def url(self, name):
        """Return the web-accessible URL for the file with filename `name`.

        settings.DATABASE_FILES_URL_METHOD(name) will return a relative URL. So we have to try to get the absolute url
        for the file. If self._base_url hasn't been set we will return the relative URL (by prepending just '').
        We could also raise an error like the SFTP backend of django-storages is doing, but the relative URL is the
        previous behavior so we fall back to that.
        """
        return "{}{}".format(self._base_url, settings.DATABASE_FILES_URL_METHOD(name))

    def size(self, name):
        """Return the size of the file with filename `name` in bytes."""
        try:
            return models.File.objects.get_from_name(name).size
        except models.File.DoesNotExist:
            return super(DatabaseStorage, self).size(name)
